@@ -1,30 +1,30 @@
+import warnings
+
 try:
-    from collections.abc import Sized
+    pass
 except ImportError:
-    from collections import Sized
-from collections import defaultdict
-from functools import partial
+    pass
 
 import numpy as np
 from scipy.stats import rankdata
-
-import sklearn
-from sklearn.base import is_classifier, clone
-from joblib import Parallel, delayed
 from sklearn.model_selection._search import BaseSearchCV
 from sklearn.utils import check_random_state
-from sklearn.utils.fixes import MaskedArray
+from sklearn.utils.validation import check_is_fitted
 
-from sklearn.utils.validation import indexable, check_is_fitted
 try:
-    from sklearn.metrics import check_scoring
+    pass
 except ImportError:
-    from sklearn.metrics.scorer import check_scoring
+    pass
 
 from . import Optimizer
-from .utils import point_asdict, dimensions_aslist, eval_callbacks
-from .space import check_dimension
 from .callbacks import check_callback
+from .space import check_dimension
+from .utils import dimensions_aslist, eval_callbacks, point_asdict
+
+
+def _get_score_names(cv_results, *, kind="test"):
+    prefix = f"mean_{kind}_"
+    return {key[len(prefix) :] for key in cv_results.keys() if key.startswith(prefix)}
 
 
 class BayesSearchCV(BaseSearchCV):
@@ -53,8 +53,7 @@ class BayesSearchCV(BaseSearchCV):
         Either estimator needs to provide a ``score`` function,
         or ``scoring`` must be passed.
 
-    search_spaces : dict, list of dict or list of tuple containing
-        (dict, int).
+    search_spaces : dict, list of dict or list of tuple containing (dict, int).
         One of these cases:
         1. dictionary, where keys are parameter names (strings)
         and values are skopt.space.Dimension instances (Real, Integer
@@ -83,11 +82,23 @@ class BayesSearchCV(BaseSearchCV):
         ``{'base_estimator': 'RF'}`` would use a Random Forest surrogate
         instead of the default Gaussian Process.
 
-    scoring : string, callable or None, default=None
-        A string (see model evaluation documentation) or
-        a scorer callable object / function with signature
-        ``scorer(estimator, X, y)``.
-        If ``None``, the ``score`` method of the estimator is used.
+    scoring : str, callable, list, tuple or dict, default=None
+        Strategy to evaluate the performance of the cross-validated model on
+        the test set. If ``None``, the ``score`` method of the estimator is
+        used.
+        If `scoring` represents a single score, one can use:
+
+        - a single string (see :ref:`scoring_parameter`);
+        - a callable (see :ref:`scoring`) that returns a single value.
+
+        If `scoring` represents multiple scores, one can use:
+
+        - a list or tuple of unique strings;
+        - a callable returning a dictionary where the keys are the metric
+          names and the values are the metric scores;
+        - a dictionary with metric names as keys and callables a values.
+
+        Callables must have the signature ``scorer(estimator, X, y=None)``
 
     fit_params : dict, optional
         Parameters to pass to the fit method.
@@ -116,11 +127,6 @@ class BayesSearchCV(BaseSearchCV):
             - A string, giving an expression as a function of n_jobs,
               as in '2*n_jobs'
 
-    iid : boolean, default=True
-        If True, the data is assumed to be identically distributed across
-        the folds, and the loss minimized is the total loss per sample,
-        and not the mean loss across the folds.
-
     cv : int, cross-validation generator or an iterable, optional
         Determines the cross-validation splitting strategy.
         Possible inputs for cv are:
@@ -134,10 +140,13 @@ class BayesSearchCV(BaseSearchCV):
         either binary or multiclass, :class:`StratifiedKFold` is used. In all
         other cases, :class:`KFold` is used.
 
-    refit : boolean, default=True
+    refit : bool, str, default=True
         Refit the best estimator with the entire dataset.
         If "False", it is impossible to make predictions using
-        this RandomizedSearchCV instance after fitting.
+        this BayesSearchCV instance after fitting.
+        For multiple metric evaluation, this needs to be a `str` denoting the
+        scorer that would be used to direct the optimization process, and find
+        the best parameters for refitting the estimator at the end.
 
     verbose : integer
         Controls the verbosity: the higher, the more messages.
@@ -167,7 +176,7 @@ class BayesSearchCV(BaseSearchCV):
     >>> from sklearn.svm import SVC
     >>> from sklearn.model_selection import train_test_split
     >>>
-    >>> X, y = load_iris(True)
+    >>> X, y = load_iris(return_X_y=True)
     >>> X_train, X_test, y_train, y_test = train_test_split(X, y,
     ...                                                     train_size=0.75,
     ...                                                     random_state=0)
@@ -181,7 +190,7 @@ class BayesSearchCV(BaseSearchCV):
     ...         'degree': Integer(1,8),
     ...         'kernel': Categorical(['linear', 'poly', 'rbf']),
     ...     },
-    ...     n_iter=32,
+    ...     n_iter=10, n_jobs=-1,
     ...     random_state=0
     ... )
     >>>
@@ -268,13 +277,19 @@ class BayesSearchCV(BaseSearchCV):
     n_splits_ : int
         The number of cross-validation splits (folds/iterations).
 
+    refit_time_ : float
+        Seconds used for refitting the best model on the whole dataset.
+        This is present only if ``refit`` is not False.
+    multimetric_ : bool
+        Whether or not the scorers compute several metrics.
+
     Notes
     -----
     The parameters selected are those that maximize the score of the held-out
     data, according to the scoring parameter.
 
     If `n_jobs` was set to a value higher than one, the data is copied for each
-    parameter setting(and not `n_jobs` times). This is done for efficiency
+    parameter setting (and not `n_jobs` times). This is done for efficiency
     reasons if individual jobs take very little time, but may raise errors if
     the dataset is large and not enough memory is available.  A workaround in
     this case is to set `pre_dispatch`. Then, the memory is copied only
@@ -285,14 +300,27 @@ class BayesSearchCV(BaseSearchCV):
     --------
     :class:`GridSearchCV`:
         Does exhaustive search over a grid of parameters.
-
     """
 
-    def __init__(self, estimator, search_spaces, optimizer_kwargs=None,
-                 n_iter=50, scoring=None, fit_params=None, n_jobs=1,
-                 n_points=1, iid=True, refit=True, cv=None, verbose=0,
-                 pre_dispatch='2*n_jobs', random_state=None,
-                 error_score='raise', return_train_score=False):
+    def __init__(
+        self,
+        estimator,
+        search_spaces,
+        optimizer_kwargs=None,
+        n_iter=50,
+        scoring=None,
+        fit_params=None,
+        n_jobs=1,
+        n_points=1,
+        iid='deprecated',
+        refit=True,
+        cv=None,
+        verbose=0,
+        pre_dispatch='2*n_jobs',
+        random_state=None,
+        error_score='raise',
+        return_train_score=False,
+    ):
 
         self.search_spaces = search_spaces
         self.n_iter = n_iter
@@ -306,14 +334,26 @@ class BayesSearchCV(BaseSearchCV):
         # in the constructor and be passed in ``fit``.
         self.fit_params = fit_params
 
-        super(BayesSearchCV, self).__init__(
-             estimator=estimator, scoring=scoring,
-             n_jobs=n_jobs, iid=iid, refit=refit, cv=cv, verbose=verbose,
-             pre_dispatch=pre_dispatch, error_score=error_score,
-             return_train_score=return_train_score)
+        if iid != "deprecated":
+            warnings.warn(
+                "The `iid` parameter has been deprecated " "and will be ignored."
+            )
+        self.iid = iid  # For sklearn repr pprint
+
+        super().__init__(
+            estimator=estimator,
+            scoring=scoring,
+            n_jobs=n_jobs,
+            refit=refit,
+            cv=cv,
+            verbose=verbose,
+            pre_dispatch=pre_dispatch,
+            error_score=error_score,
+            return_train_score=return_train_score,
+        )
 
     def _check_search_space(self, search_space):
-        """Checks whether the search space argument is correct"""
+        """Checks whether the search space argument is correct."""
 
         if len(search_space) == 0:
             raise ValueError(
@@ -343,8 +383,7 @@ class BayesSearchCV(BaseSearchCV):
                     if (not isinstance(n_iter, int)) or n_iter < 0:
                         raise ValueError(
                             "Number of iterations in search space should be"
-                            "positive integer, got %s in tuple %s " %
-                            (n_iter, elem)
+                            "positive integer, got %s in tuple %s " % (n_iter, elem)
                         )
 
                     # save subspaces here for further checking
@@ -354,181 +393,24 @@ class BayesSearchCV(BaseSearchCV):
                 else:
                     raise TypeError(
                         "A search space should be provided as a dict or"
-                        "tuple (dict, int), got %s" % elem)
+                        "tuple (dict, int), got %s" % elem
+                    )
 
             # 2. check all the dicts for correctness of contents
             for subspace in dicts_only:
-                for k, v in subspace.items():
+                for _, v in subspace.items():
                     check_dimension(v)
         else:
             raise TypeError(
                 "Search space should be provided as a dict or list of dict,"
-                "got %s" % search_space)
-
-    # copied for compatibility with 0.19 sklearn from 0.18 BaseSearchCV
-    @property
-    def best_score_(self):
-        check_is_fitted(self, 'cv_results_')
-        return self.cv_results_['mean_test_score'][self.best_index_]
-
-    # copied for compatibility with 0.19 sklearn from 0.18 BaseSearchCV
-    @property
-    def best_params_(self):
-        check_is_fitted(self, 'cv_results_')
-        return self.cv_results_['params'][self.best_index_]
+                "got %s" % search_space
+            )
 
     @property
     def optimizer_results_(self):
         check_is_fitted(self, '_optim_results')
         return self._optim_results
 
-    # copied for compatibility with 0.19 sklearn from 0.18 BaseSearchCV
-    def _fit(self, X, y, groups, parameter_iterable):
-        """
-        Actual fitting,  performing the search over parameters.
-        Taken from https://github.com/scikit-learn/scikit-learn/blob/0.18.X
-                    .../sklearn/model_selection/_search.py
-        """
-        estimator = self.estimator
-        cv = sklearn.model_selection._validation.check_cv(
-            self.cv, y, classifier=is_classifier(estimator))
-        self.scorer_ = check_scoring(
-            self.estimator, scoring=self.scoring)
-
-        X, y, groups = indexable(X, y, groups)
-        n_splits = cv.get_n_splits(X, y, groups)
-        if self.verbose > 0 and isinstance(parameter_iterable, Sized):
-            n_candidates = len(parameter_iterable)
-            print("Fitting {0} folds for each of {1} candidates, totalling"
-                  " {2} fits".format(n_splits, n_candidates,
-                                     n_candidates * n_splits))
-
-        base_estimator = clone(self.estimator)
-        pre_dispatch = self.pre_dispatch
-
-        cv_iter = list(cv.split(X, y, groups))
-        out = Parallel(
-            n_jobs=self.n_jobs, verbose=self.verbose,
-            pre_dispatch=pre_dispatch
-        )(delayed(sklearn.model_selection._validation._fit_and_score)(
-                clone(base_estimator),
-                X, y, self.scorer_,
-                train, test, self.verbose, parameters,
-                fit_params=self.fit_params,
-                return_train_score=self.return_train_score,
-                return_n_test_samples=True,
-                return_times=True, return_parameters=True,
-                error_score=self.error_score
-            )
-            for parameters in parameter_iterable
-            for train, test in cv_iter)
-
-        # if one choose to see train score, "out" will contain train score info
-        if self.return_train_score:
-            (train_scores, test_scores, test_sample_counts,
-             fit_time, score_time, parameters) = zip(*out)
-        else:
-            (test_scores, test_sample_counts,
-             fit_time, score_time, parameters) = zip(*out)
-
-        candidate_params = parameters[::n_splits]
-        n_candidates = len(candidate_params)
-
-        results = dict()
-
-        def _store(key_name, array, weights=None, splits=False, rank=False):
-            """A small helper to store the scores/times to the cv_results_"""
-            array = np.array(array, dtype=np.float64).reshape(n_candidates,
-                                                              n_splits)
-            if splits:
-                for split_i in range(n_splits):
-                    results["split%d_%s"
-                            % (split_i, key_name)] = array[:, split_i]
-
-            array_means = np.average(array, axis=1, weights=weights)
-            results['mean_%s' % key_name] = array_means
-            # Weighted std is not directly available in numpy
-            array_stds = np.sqrt(np.average((array -
-                                             array_means[:, np.newaxis]) ** 2,
-                                            axis=1, weights=weights))
-            results['std_%s' % key_name] = array_stds
-
-            if rank:
-                results["rank_%s" % key_name] = np.asarray(
-                    rankdata(-array_means, method='min'), dtype=np.int32)
-
-        # Computed the (weighted) mean and std for test scores alone
-        # NOTE test_sample counts (weights) remain the same for all candidates
-        test_sample_counts = np.array(test_sample_counts[:n_splits],
-                                      dtype=np.int)
-
-        _store('test_score', test_scores, splits=True, rank=True,
-               weights=test_sample_counts if self.iid else None)
-        if self.return_train_score:
-            _store('train_score', train_scores, splits=True)
-        _store('fit_time', fit_time)
-        _store('score_time', score_time)
-
-        best_index = np.flatnonzero(results["rank_test_score"] == 1)[0]
-        best_parameters = candidate_params[best_index]
-
-        # Use one MaskedArray and mask all the places where the param is not
-        # applicable for that candidate. Use defaultdict as each candidate may
-        # not contain all the params
-        param_results = defaultdict(partial(np.ma.array,
-                                            np.empty(n_candidates,),
-                                            mask=True,
-                                            dtype=object))
-        for cand_i, params in enumerate(candidate_params):
-            for name, value in params.items():
-                # An all masked empty array gets created for the key
-                # `"param_%s" % name` at the first occurence of `name`.
-                # Setting the value at an index also unmasks that index
-                param_results["param_%s" % name][cand_i] = value
-
-        results.update(param_results)
-
-        # Store a list of param dicts at the key 'params'
-        results['params'] = candidate_params
-
-        self.cv_results_ = results
-        self.best_index_ = best_index
-        self.n_splits_ = n_splits
-
-        if self.refit:
-            # fit the best estimator using the entire dataset
-            # clone first to work around broken estimators
-            best_estimator = clone(base_estimator).set_params(
-                **best_parameters)
-            if y is not None:
-                best_estimator.fit(X, y, **self.fit_params)
-            else:
-                best_estimator.fit(X, **self.fit_params)
-            self.best_estimator_ = best_estimator
-        return self
-
-    def _fit_best_model(self, X, y):
-        """Fit the estimator copy with best parameters found to the
-        provided data.
-
-        Parameters
-        ----------
-        X : array-like, shape = [n_samples, n_features]
-            Input data, where n_samples is the number of samples and
-            n_features is the number of features.
-
-        y : array-like, shape = [n_samples] or [n_samples, n_output],
-            Target relative to X for classification or regression.
-
-        Returns
-        -------
-        self
-        """
-        self.best_estimator_ = clone(self.estimator)
-        self.best_estimator_.set_params(**self.best_params_)
-        self.best_estimator_.fit(X, y, **(self.fit_params or {}))
-        return self
-
     def _make_optimizer(self, params_space):
         """Instantiate skopt Optimizer class.
 
@@ -543,7 +425,6 @@ class BayesSearchCV(BaseSearchCV):
         -------
         optimizer: Instance of the `Optimizer` class used for for search
             in some parameter space.
-
         """
 
         kwargs = self.optimizer_kwargs_.copy()
@@ -552,15 +433,14 @@ class BayesSearchCV(BaseSearchCV):
         for i in range(len(optimizer.space.dimensions)):
             if optimizer.space.dimensions[i].name is not None:
                 continue
-            optimizer.space.dimensions[i].name = list(sorted(
-                params_space.keys()))[i]
+            optimizer.space.dimensions[i].name = list(sorted(params_space.keys()))[i]
 
         return optimizer
 
-    def _step(self, X, y, search_space, optimizer, groups=None, n_points=1):
-        """Generate n_jobs parameters and evaluate them in parallel.
-        """
-
+    def _step(
+        self, search_space, optimizer, score_name, evaluate_candidates, n_points=1
+    ):
+        """Generate n_jobs parameters and evaluate them in parallel."""
         # get parameter values to evaluate
         params = optimizer.ask(n_points=n_points)
 
@@ -570,40 +450,41 @@ class BayesSearchCV(BaseSearchCV):
         # make lists into dictionaries
         params_dict = [point_asdict(search_space, p) for p in params]
 
-        # HACK: self.cv_results_ is reset at every call to _fit, keep current
-        all_cv_results = self.cv_results_
+        all_results = evaluate_candidates(params_dict)
 
-        # HACK: this adds compatibility with different versions of sklearn
-        refit = self.refit
-        self.refit = False
-        self._fit(X, y, groups, params_dict)
-        self.refit = refit
+        # if self.scoring is a callable, we have to wait until here
+        # to get the score name
+        if score_name is None:
+            score_names = _get_score_names(all_results)
+            if len(score_names) > 1:
+                # multimetric case
+                # early check to fail before lengthy computations, as
+                # BaseSearchCV only performs this check *after* _run_search
+                self._check_refit_for_multimetric(score_names)
+                score_name = f"mean_test_{self.refit}"
+            elif len(score_names) == 1:
+                # single metric, or a callable self.scoring returning a dict
+                # with a single value
+                # In both case, we just use the score that is available
+                score_name = f"mean_test_{score_names.pop()}"
+            else:
+                # failsafe, shouldn't happen
+                raise ValueError(
+                    "No score was detected after fitting. This is probably "
+                    "due to a callable 'scoring' returning an empty dict."
+                )
 
-        # merge existing and new cv_results_
-        for k in self.cv_results_:
-            all_cv_results[k].extend(self.cv_results_[k])
-
-        all_cv_results["rank_test_score"] = list(np.asarray(
-            rankdata(-np.array(all_cv_results['mean_test_score']),
-                     method='min'), dtype=np.int32))
-        if self.return_train_score:
-            all_cv_results["rank_train_score"] = list(np.asarray(
-                rankdata(-np.array(all_cv_results['mean_train_score']),
-                         method='min'), dtype=np.int32))
-        self.cv_results_ = all_cv_results
-        self.best_index_ = np.argmax(self.cv_results_['mean_test_score'])
-
-        # feed the point and objective back into optimizer
-        local_results = self.cv_results_['mean_test_score'][-len(params):]
-
-        # optimizer minimizes objective, hence provide negative score
-        return optimizer.tell(params, [-score for score in local_results])
+        # Feed the point and objective value back into optimizer
+        # Optimizer minimizes objective, hence provide negative score
+        local_results = all_results[score_name][-len(params) :]
+        # return the score_name to cache it if callable refit
+        # this avoids checking self.refit all the time
+        return (optimizer.tell(params, [-score for score in local_results]), score_name)
 
     @property
     def total_iterations(self):
-        """
-        Count total iterations that will be taken to explore
-        all subspaces with `fit` method.
+        """Count total iterations that will be taken to explore all subspaces with `fit`
+        method.
 
         Returns
         -------
@@ -622,10 +503,8 @@ class BayesSearchCV(BaseSearchCV):
 
         return total_iter
 
-    def _run_search(self, x):
-        pass
-
-    def fit(self, X, y=None, groups=None, callback=None):
+    # TODO: Accept callbacks via the constructor?
+    def fit(self, X, y=None, *, groups=None, callback=None, **fit_params):
         """Run fit on the estimator with randomly drawn parameters.
 
         Parameters
@@ -646,21 +525,52 @@ class BayesSearchCV(BaseSearchCV):
             combination tested. If list of callables, then each callable in
             the list is called.
         """
-
-        # check if space is a single dict, convert to list if so
-        search_spaces = self.search_spaces
-        if isinstance(search_spaces, dict):
-            search_spaces = [search_spaces]
-
-        callbacks = check_callback(callback)
+        self._callbacks = check_callback(callback)
 
         if self.optimizer_kwargs is None:
             self.optimizer_kwargs_ = {}
         else:
             self.optimizer_kwargs_ = dict(self.optimizer_kwargs)
+
+        if callable(self.refit):
+            raise ValueError(
+                "BayesSearchCV doesn't support a callable refit, "
+                "as it doesn't define an implicit score to "
+                "optimize"
+            )
+
+        super().fit(X=X, y=y, groups=groups, **fit_params)
+
+        # BaseSearchCV never ranked train scores,
+        # but apparently we used to ship this (back-compat)
+        if self.return_train_score:
+            for score in _get_score_names(self.cv_results_, kind="train"):
+                self.cv_results_[f"rank_train_{score}"] = rankdata(
+                    -np.array(self.cv_results_[f"mean_train_{score}"]), method='min'
+                ).astype(int)
+        return self
+
+    def _run_search(self, evaluate_candidates):
+        # check if space is a single dict, convert to list if so
+        search_spaces = self.search_spaces
+        if isinstance(search_spaces, dict):
+            search_spaces = [search_spaces]
+
+        callbacks = self._callbacks
+
         random_state = check_random_state(self.random_state)
         self.optimizer_kwargs_['random_state'] = random_state
 
+        # Adapted from BaseSearchCV fit() method
+        if callable(self.scoring):
+            # will be determined later
+            score_name = None
+        elif self.scoring is None or isinstance(self.scoring, str):
+            score_name = "mean_test_score"
+        else:
+            # proper checking took place before in BaseSearchCV.fit()
+            score_name = f"mean_test_{self.refit}"
+
         # Instantiate optimizers for all the search spaces.
         optimizers = []
         for search_space in search_spaces:
@@ -669,9 +579,6 @@ class BayesSearchCV(BaseSearchCV):
             optimizers.append(self._make_optimizer(search_space))
         self.optimizers_ = optimizers  # will save the states of the optimizers
 
-        self.cv_results_ = defaultdict(list)
-        self.best_index_ = None
-        self.multimetric_ = False
         self._optim_results = []
 
         n_points = self.n_points
@@ -689,9 +596,12 @@ class BayesSearchCV(BaseSearchCV):
                 # when n_iter < n_points points left for evaluation
                 n_points_adjusted = min(n_iter, n_points)
 
-                optim_result = self._step(
-                    X, y, search_space, optimizer,
-                    groups=groups, n_points=n_points_adjusted
+                optim_result, score_name = self._step(
+                    search_space,
+                    optimizer,
+                    score_name,
+                    evaluate_candidates,
+                    n_points=n_points_adjusted,
                 )
                 n_iter -= n_points
 
@@ -699,8 +609,18 @@ class BayesSearchCV(BaseSearchCV):
                     break
             self._optim_results.append(optim_result)
 
-        # Refit the best model on the the whole dataset
-        if self.refit:
-            self._fit_best_model(X, y)
+    def _check_refit_for_multimetric(self, scores):
+        """Check `refit` is compatible with `scores` and valid."""
+        # override parent method to exclude False and callables
+        multimetric_refit_msg = (
+            "For multi-metric scoring, the 'refit' parameter must be set to a "
+            "scorer key, used to guide the bayesian optimization process "
+            "and refit an estimator with the best parameter settings on the "
+            "whole dataset (making the best_* attributes available for that "
+            f" metric). {self.refit!r} was passed."
+        )
 
-        return self
+        is_refit_valid = isinstance(self.refit, str) and self.refit in scores
+
+        if not is_refit_valid:
+            raise ValueError(multimetric_refit_msg)
